# General packages
library(dplyr)
library(tidyr)
library(furrr)

# Geo packages
library(terra)
library(CoordinateCleaner)
library(sf)

# DB packages
library(Symobio)
library(DBI)

source("R/utils.R")

con = db_connect() # Connection to Symobio
sf::sf_use_s2(use_s2 = FALSE) # Don't use spherical geometry

# ---------------------------------------------------------------------------#
# Prepare Geodata                                                         ####
# ---------------------------------------------------------------------------#
raster_info = tbl(con, "datasets") %>% 
  dplyr::filter(stringr::str_detect(name, "CHELSA")) %>% 
  collect()

# raster_filepaths = list.files(raster_info$raw_data_path, pattern = ".tif$", full.names = T) %>% 
#   stringr::str_sort(numeric = T)

raster_filepaths = list.files("I:/mas_data/00_data/processed/CHELSA_v2-1_bioclim", pattern = ".tif$", full.names = T) %>% 
  stringr::str_sort(numeric = T)

sa_polygon = rnaturalearth::ne_countries() %>% 
  dplyr::filter(continent == "South America") %>% 
  sf::st_union()

# ---------------------------------------------------------------------------#
# Prepare Occurrence Data                                                 ####
# ---------------------------------------------------------------------------#
load("data/r_objects/range_maps.RData")
target_species = unique(range_maps$name_matched[!is.na(range_maps$name_matched)])

occs = tbl(con, "species_occurrences") %>% 
  dplyr::filter(species %in% target_species) %>% 
  dplyr::select(-year) %>% 
  dplyr::distinct() %>% 
  collect() %>% 
  sf::st_as_sf(coords = c("longitude", "latitude"), remove = F, crs = sf::st_crs(4326)) %>% 
  sf::st_filter(sa_polygon)

occs_flagged = occs %>% 
  dplyr::distinct(species, coordinate_id, longitude, latitude) %>% 
  group_by(species) %>% 
  group_split() %>% 
  purrr::map(                     # Loop over species individually due to bug in CoordinateCleaner
    CoordinateCleaner::clean_coordinates,
    lon = "longitude", 
    lat = "latitude",
    tests = c("centroids", "gbif", "equal", "institutions", "outliers"),
    outliers_method = "quantile",
    verbose = F
  ) %>% 
  bind_rows() %>% 
  dplyr::filter(.summary == T) %>% 
  dplyr::select(species, coordinate_id, longitude, latitude)

env_vars = tbl(con, "raster_extracts_num") %>% 
  dplyr::filter(
    coordinate_id %in% occs$coordinate_id,
    metric == "mean"
  ) %>% 
  arrange(raster_layer_id) %>% 
  tidyr::pivot_wider(id_cols = coordinate_id, names_from = raster_layer_id, names_prefix = "layer_", values_from = value) %>% 
  collect()

occs_final = occs %>% 
  inner_join(occs_flagged, by = c("species", "coordinate_id", "longitude", "latitude")) %>% 
  inner_join(env_vars, by = "coordinate_id") %>% 
  dplyr::select(-coordinate_id)

save(occs_final, file = "data/r_objects/occs_final.RData")

# ---------------------------------------------------------------------------#
# Create SSDM dataset                                                     ####
# ---------------------------------------------------------------------------#
load("data/r_objects/occs_final.RData")
load("data/r_objects/sa_polygon.RData")
sf::sf_use_s2(use_s2 = FALSE)

occs_split = split(occs_final, occs_final$species)

future::plan("multisession", workers = 8)
ssdm_data = furrr::future_map(occs_split, .progress = TRUE, .options = furrr::furrr_options(seed = 42), .f = function(occs_spec){
  # skip low information species
  if(nrow(occs_spec) < 5){
    return(NULL)
  }
  
  # Define model/sampling region
  occs_bbox = occs_spec %>% 
    sf::st_bbox() %>% 
    expand_bbox(min_span = 1, expansion = 0.25) %>% 
    sf::st_as_sfc() %>% 
    st_set_crs(st_crs(occs_spec)) 
  
  sample_region = suppressMessages(
    st_as_sf(st_intersection(occs_bbox, sa_polygon))
  )
  
  # Sample pseudo absence
  sample_points = st_as_sf(st_sample(sample_region, nrow(occs_spec))) %>% 
    dplyr::mutate(
      species = occs_spec$species[1],
      longitude = sf::st_coordinates(.)[,1],
      latitude = sf::st_coordinates(.)[,2]
    ) %>% 
    dplyr::rename(geometry = "x")
  
  abs_spec = terra::rast(raster_filepaths) %>% 
    setNames(paste0("layer_", 1:19)) %>% 
    terra::extract(sample_points) %>% 
    dplyr::select(-ID) %>% 
    dplyr::mutate(
      present = 0,
      geometry = sample_points$x
    ) %>% 
    tibble() %>% 
    bind_cols(sample_points)
  
  # Create presence-absence dataframe
  pa_spec = occs_spec %>% 
    dplyr::mutate(present = 1) %>% 
    bind_rows(abs_spec) 
  
  # Define cross-validation folds
  spatial_folds = blockCV::cv_spatial(
    pa_spec,
    column = "present",
    k = 5,
    progress = F, plot = F, report = F
  )
  
  pa_spec$fold = spatial_folds$folds_ids
  pa_spec$geometry = NULL
  
  # Split into train and test datasets
  train_index = createDataPartition(pa_spec$present, p = 0.7, list = FALSE)
  pa_spec$train = 0
  pa_spec$train[train_index] = 1
  
  return(pa_spec)
})

ssdm_data = bind_rows(ssdm_data)
save(ssdm_data, file = "data/r_objects/ssdm_data.RData")

# ---------------------------------------------------------------------------#
# Create MSDM dataset                                                     ####
# ---------------------------------------------------------------------------#
load("data/r_objects/occs_final.RData")
load("data/r_objects/sa_polygon.RData")
sf::sf_use_s2(use_s2 = FALSE)

occs_split = split(occs_final, occs_final$species)

future::plan("multisession", workers = 8)
msdm_data = furrr::future_map(occs_split, .progress = TRUE, .options = furrr::furrr_options(seed = 42), .f = function(occs_spec){
  # Define model/sampling region
  occs_bbox = occs_spec %>% 
    sf::st_bbox() %>% 
    expand_bbox(min_span = 1, expansion = 0.25) %>% 
    sf::st_as_sfc() %>% 
    st_set_crs(st_crs(occs_spec)) 
  
  sample_region = suppressMessages(
    st_as_sf(st_intersection(occs_bbox, sa_polygon))
  )
  
  # Sample pseudo absence
  sample_points = st_as_sf(st_sample(sample_region, nrow(occs_spec))) %>% 
    dplyr::mutate(
      species = occs_spec$species[1],
      longitude = sf::st_coordinates(.)[,1],
      latitude = sf::st_coordinates(.)[,2]
    ) %>% 
    dplyr::rename(geometry = "x")
  
  abs_spec = terra::rast(raster_filepaths) %>% 
    setNames(paste0("layer_", 1:19)) %>% 
    terra::extract(sample_points) %>% 
    dplyr::select(-ID) %>% 
    dplyr::mutate(
      present = 0,
      geometry = sample_points$x
    ) %>% 
    tibble() %>% 
    bind_cols(sample_points)
  
  # Create presence-absence dataframe
  pa_spec = occs_spec %>% 
    dplyr::mutate(present = 1) %>% 
    bind_rows(abs_spec) 
  
  return(pa_spec)
})

msdm_data = bind_rows(msdm_data)
save(ssdm_data, file = "data/r_objects/msdm_data.RData")
