# ---------------------------------------------------------------------------#
# Main loop                                                               ####
# ---------------------------------------------------------------------------#
load("data/r_objects/occs_final.RData")
load("data/r_objects/sa_polygon.RData")

occs_split = split(occs_final, occs_final$species)

future::plan("multisession", workers = 20)

ssdm_data = furrr::future_map(occs_split[1:20], .options = furrr::furrr_options(seed = 123), .f = function(occs_spec){
  # Initial check
  if(nrow(occs_spec) < 10){
    return(NULL)
  }
  
  species = occs_spec$species[1]
  
  # ------------------------------- #
  # Define model/sampling region ####
  # ------------------------------- #
  sample_region = tryCatch({
    sf::sf_use_s2(use_s2 = FALSE)
    
    occs_bbox = occs_spec %>% 
      sf::st_bbox() %>% 
      expand_bbox(0.25) %>% 
      sf::st_as_sfc() %>% 
      st_set_crs(st_crs(occs_spec)) 
    
    sample_region = suppressMessages(
      st_as_sf(st_intersection(occs_bbox, sa_polygon))
    )
  }, error = function(e){
    glue::glue("Skipping species {species}:\n{e}")
  })
  
  if(is.character(sample_region)){ # There was an error
    return(sample_region)
  }
  
  # ------------------------------- #
  # Create pseudo absence        ####
  # ------------------------------- #
  sample_points = st_as_sf(st_sample(sample_region, nrow(occs_spec)))  # TODO: Vary sample size?
  
  abs_spec = terra::rast(raster_filepaths) %>% 
    setNames(paste0("layer_", 1:19)) %>% 
    terra::extract(sample_points) %>% 
    dplyr::select(-ID) %>% 
    dplyr::mutate(
      presence = 0,
      geometry = sample_points$x
    ) %>% 
    tibble()
  
  # ------------------------------- #
  # Create modeling dataset      ####
  # ------------------------------- #
  model_data = occs_spec %>% 
    dplyr::mutate(presence = 1) %>% 
    bind_rows(abs_spec) %>% 
    dplyr::mutate(
      presence_fct = as.factor(ifelse(presence == 0, "absent", "present"))
    )
  
  # Define cross-validation folds
  spatial_folds = blockCV::cv_spatial(
    model_data,
    column = "presence",
    k = 5,
    progress = F, plot = F, report = F
  )
  
  model_data$fold = spatial_folds$folds_ids
  model_data$geometry = NULL
  
  # Split into train and test datasets
  train_index = createDataPartition(model_data$presence, p = 0.7, list = FALSE)
  train_data = model_data[train_index, ]
  test_data  = model_data[-train_index, ]
  
  # Define predictor columns
  predictors = paste0("layer_", 1:19)
  
  # Define empty performance summary
  na_performance = list(    
    AUC = NA,
    Accuracy = NA,
    Kappa = NA,
    Precision = NA,
    Recall = NA,
    F1 = NA
  )
  
  # ------------------------------- #
  # Train models                 ####
  # ------------------------------- #
  ## cito ####
  config_lr_scheduler()
  
  nn_performance = tryCatch({
    nn_fit = dnn(
      X = train_data[, predictors],
      Y = train_data$presence,
      hidden = c(500L, 200L, 50L),
      loss = "binomial",
      activation = c("sigmoid", "leaky_relu", "leaky_relu"),
      epochs = 500L, 
      lr = 0.02,   
      baseloss=10,
      batchsize=nrow(train_data)/4,
      dropout = 0.1,  # Regularization 
      optimizer = config_optimizer("adam", weight_decay = 0.001),
      lr_scheduler = config_lr_scheduler("reduce_on_plateau", patience = 100, factor = 0.7),
      early_stopping = 200, # stop training when validation loss does not decrease anymore
      validation = 0.3, # used for early stopping and lr_scheduler 
      device = "cuda",
      bootstrap = 5
    )
    
    evaluate_model(nn_fit, test_data)
  }, error = function(e){
    na_performance
  })
  
  ## caret ####  
  # Define training routine
  index_train = lapply(unique(sort(train_data$fold)), function(x){
    return(which(train_data$fold != x))
  })
  
  train_ctrl = trainControl(
    search = "grid",
    classProbs = TRUE, 
    index = index_train,
    summaryFunction = twoClassSummary, 
    savePredictions = "final"
  )
  
  # Random Forest
  rf_performance = tryCatch({
    rf_grid = expand.grid(
      mtry = c(3,7,11,15,19)                # Number of randomly selected predictors
    )
    
    rf_fit = caret::train(
      x = train_data[, predictors],
      y = train_data$presence_fct,
      method = "rf",
      metric = "ROC",
      tuneGrid = rf_grid,
      trControl = train_ctrl
    )
    evaluate_model(rf_fit, test_data)
  }, error = function(e){
    na_performance
  })
  
  # Gradient Boosted Machine
  gbm_performance = tryCatch({
    gbm_grid <- expand.grid(
      n.trees = c(100, 500, 1000, 1500),       # Higher number of boosting iterations
      interaction.depth = c(3, 5, 7),          # Maximum depth of each tree
      shrinkage = c(0.01, 0.005, 0.001),       # Lower learning rates
      n.minobsinnode = c(10, 20)               # Minimum number of observations in nodes
    )
    
    gbm_fit = train(
      x = train_data[, predictors],
      y = train_data$presence_fct,
      method = "gbm",
      metric = "ROC",
      verbose = F,
      tuneGrid = gbm_grid,
      trControl = train_ctrl
    )
    evaluate_model(gbm_fit, test_data)
  }, error = function(e){
    na_performance
  })
  
  # Generalized additive Model
  glm_performance = tryCatch({
    glm_fit = train(
      x = train_data[, predictors],
      y = train_data$presence_fct,
      method = "glm",
      family=binomial, 
      metric = "ROC",
      preProcess = c("center", "scale"),
      trControl = train_ctrl
    )
    evaluate_model(glm_fit, test_data)
  }, error = function(e){
    na_performance
  })
  
  # Summarize results
  performance_summary = tibble(
    species = species,
    obs = nrow(occs_spec),
    model = c("NN", "RF", "GBM", "GLM"),
    auc = c(nn_performance$AUC, rf_performance$AUC, gbm_performance$AUC, glm_performance$AUC),
    accuracy = c(nn_performance$Accuracy, rf_performance$Accuracy, gbm_performance$Accuracy, glm_performance$Accuracy),
    kappa = c(nn_performance$Kappa, rf_performance$Kappa, gbm_performance$Kappa, glm_performance$Kappa),
    precision = c(nn_performance$Precision, rf_performance$Precision, gbm_performance$Precision, glm_performance$Precision),
    recall = c(nn_performance$Recall, rf_performance$Recall, gbm_performance$Recall, glm_performance$Recall),
    f1 = c(nn_performance$F1, rf_performance$F1, gbm_performance$F1, glm_performance$F1)
  )
  
  return(performance_summary)
})

model_data_df = bind_rows(model_data)

save(model_results, file = "data/r_objects/model_results.RData")